package gnu.crypto.sig.rsa;

// ----------------------------------------------------------------------------
// $Id: RSA.java,v 1.5 2003/04/28 10:59:11 raif Exp $
//
// Copyright (C) 2001, 2002, 2003 Free Software Foundation, Inc.
//
// This file is part of GNU Crypto.
//
// GNU Crypto is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// GNU Crypto is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; see the file COPYING.  If not, write to the
//
//    Free Software Foundation Inc.,
//    59 Temple Place - Suite 330,
//    Boston, MA 02111-1307
//    USA
//
// Linking this library statically or dynamically with other modules is
// making a combined work based on this library.  Thus, the terms and
// conditions of the GNU General Public License cover the whole
// combination.
//
// As a special exception, the copyright holders of this library give
// you permission to link this library with independent modules to
// produce an executable, regardless of the license terms of these
// independent modules, and to copy and distribute the resulting
// executable under terms of your choice, provided that you also meet,
// for each linked independent module, the terms and conditions of the
// license of that module.  An independent module is a module which is
// not derived from or based on this library.  If you modify this
// library, you may extend this exception to your version of the
// library, but you are not obligated to do so.  If you do not wish to
// do so, delete this exception statement from your version.
// ----------------------------------------------------------------------------

import java.math.BigInteger;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.interfaces.RSAPrivateCrtKey;
import java.security.interfaces.RSAPrivateKey;
import java.security.interfaces.RSAPublicKey;

/**
 * <p>Utility methods related to the RSA algorithm.</p>
 *
 * <p>References:</p>
 * <ol>
 *    <li><a href="http://www.cosic.esat.kuleuven.ac.be/nessie/workshop/submissions/rsa-pss.zip">
 *    RSA-PSS Signature Scheme with Appendix, part B.</a><br>
 *    Primitive specification and supporting documentation.<br>
 *    Jakob Jonsson and Burt Kaliski.</li>
 *
 *    <li><a href="http://www.ietf.org/rfc/rfc3447.txt">Public-Key Cryptography
 *    Standards (PKCS) #1:</a><br>
 *    RSA Cryptography Specifications Version 2.1.<br>
 *    Jakob Jonsson and Burt Kaliski.</li>
 * </ol>
 *
 * @version $Revision: 1.5 $
 */
public class RSA {

   // Constants and variables
   // -------------------------------------------------------------------------

   private static final BigInteger ZERO = BigInteger.ZERO;
   private static final BigInteger ONE = BigInteger.ONE;

   // Constructor(s)
   // -------------------------------------------------------------------------

   /** Trivial private constructor to enforce Singleton pattern. */
   private RSA() {
      super();
   }

   // Class methods
   // -------------------------------------------------------------------------

   /**
    * <p>An implementation of the <b>RSASP</b> method: Assuming that the
    * designated RSA private key is a valid one, this method computes a
    * <i>signature representative</i> for a designated <i>message
    * representative</i> signed by the holder of the designated RSA private
    * key.<p>
    *
    * @param K the RSA private key.
    * @param m the <i>message representative</i>: an integer between
    * <code>0</code> and <code>n - 1</code>, where <code>n</code> is the RSA
    * <i>modulus</i>.
    * @return the <i>signature representative</i>, an integer between
    * <code>0</code> and <code>n - 1</code>, where <code>n</code> is the RSA
    * <i>modulus</i>.
    * @throws IllegalArgumentException if the key is not an RSA one, or if
    * <code>m</code> (the <i>message representative</i>) is out of range.
    */
   public static final BigInteger sign(PrivateKey K, BigInteger m) {
      if (!(K instanceof RSAPrivateKey)) {
         throw new IllegalArgumentException("invalid key");
      }

      // 1. If the message representative m is not between 0 and n-1,
      // output 'message representative out of range' and stop
      BigInteger n = ((RSAPrivateKey) K).getModulus();
      if (m.compareTo(ZERO) < 0 || m.compareTo(n.subtract(ONE)) > 0) {
         throw new IllegalArgumentException("message representative out of range");
      }

      BigInteger result = null;
      if (!(K instanceof RSAPrivateCrtKey)) {
         BigInteger d = ((RSAPrivateKey) K).getPrivateExponent();

         // 2. If the first form (n, d) of K is used:
         // 2.1 Let s = m**d mod n
         result = m.modPow(d, n);
      } else {
         BigInteger p =    ((RSAPrivateCrtKey) K).getPrimeP();
         BigInteger q =    ((RSAPrivateCrtKey) K).getPrimeQ();
         BigInteger dP =   ((RSAPrivateCrtKey) K).getPrimeExponentP();
         BigInteger dQ =   ((RSAPrivateCrtKey) K).getPrimeExponentQ();
         BigInteger qInv = ((RSAPrivateCrtKey) K).getCrtCoefficient();

         // Else if the second form (p, q, dP, dQ, qInv) of K is used:
         // 2.2 Let s1 = m**dP mod p
         BigInteger s1 = m.modPow(dP, p);
         // 2.3 Let s2 = m**dQ mod q
         BigInteger s2 = m.modPow(dQ, q);
         // 2.4 Let h = (s1 - s2) * qInv mod p
         BigInteger h = s1.subtract(s2).multiply(qInv).mod(p);
         // 2.5 Let s = s2 + q * h
         result = s2.add(q.multiply(h));
      }
      // 3. Output s
      return result;
   }

   /**
    * <p>An implementation of the <b>RSAVP</b> method: Assuming that the
    * designated RSA public key is a valid one, this method computes a
    * <i>message representative</i> for the designated <i>signature
    * representative</i> generated by an RSA private key, for a message
    * intended for the holder of the designated RSA public key.</p>
    *
    * @param K the RSA public key.
    * @param s the <i>signature representative</i>, an integer between
    * <code>0</code> and <code>n - 1</code>, where <code>n</code> is the RSA
    * <i>modulus</i>.
    * @return a <i>message representative</i>: an integer between <code>0</code>
    * and <code>n - 1</code>, where <code>n</code> is the RSA <i>modulus</i>.
    * @throws IllegalArgumentException if the key is not an RSA one or if
    * <code>s</code> (the <i>signature representative</i>) is out of range.
    */
   public static final BigInteger verify(PublicKey K, BigInteger s) {
      if (!(K instanceof RSAPublicKey)) {
         throw new IllegalArgumentException("wrong key");
      }

      BigInteger n = ((RSAPublicKey) K).getModulus();
      BigInteger e = ((RSAPublicKey) K).getPublicExponent();

      // 1. If the signature representative s is not between 0 and n-1,
      // output 'signature representative out of range' and stop.
      if (s.compareTo(ZERO) < 0 || s.compareTo(n.subtract(ONE)) > 0) {
         throw new IllegalArgumentException("signature representative out of range");
      }
      // 2. Let m = s**e mod n.
      BigInteger result = s.modPow(e, n);
      // 3. Output m.
      return result;
   }

   /**
    * <p>Converts a <i>multi-precision integer</i> (MPI) <code>s</code> into an
    * octet sequence of length <code>k</code>.</p>
    *
    * @param s the multi-precision integer to convert.
    * @param k the length of the output.
    * @return the result of the transform.
    * @exception IllegalArgumentException if the length in octets of meaningful
    * bytes of <code>s</code> is greater than <code>k</code>.
    */
   public static final byte[] I2OSP(BigInteger s, int k) {
      byte[] result = s.toByteArray();
      if (result.length < k) {
         byte[] newResult = new byte[k];
         System.arraycopy(result, 0, newResult, k-result.length, result.length);
         result = newResult;
      } else if (result.length > k) { // leftmost extra bytes should all be 0
         int limit = result.length - k;
         for (int i = 0; i < limit; i++) {
            if (result[i] != 0x00) {
               throw new IllegalArgumentException("integer too large");
            }
         }
         byte[] newResult = new byte[k];
         System.arraycopy(result, limit, newResult, 0, k);
         result = newResult;
      }
      return result;
   }
}
